from dataclasses import dataclass
from datetime import datetime

from psycopg import sql
from psycopg.rows import class_row
from psycopg_pool import AsyncConnectionPool
from pydantic import constr


@dataclass
class Todo:
    complete: bool
    due: datetime | None
    id: int
    task: constr(strip_whitespace=True, min_length=1)  # type: ignore


async def select_todos(
    db: AsyncConnectionPool, member_id: int, complete: bool | None = None
) -> list[Todo]:
    """
    Select multiple Todo(s) from a member_id

    Return a list of Todos (can be empty list)
    """
    async with db.connection() as conn:
        async with conn.cursor(row_factory=class_row(Todo)) as curr:

            if complete is None:
                condition = sql.SQL("member_id = {}").format(sql.Literal(member_id))
            else:
                condition = sql.SQL("member_id = {} AND complete = {}").format(
                    sql.Literal(member_id), sql.Literal(complete)
                )

            query = sql.SQL(
                """
                SELECT id, complete, due, task
                    FROM todos
                    WHERE {}
                """
            ).format(condition)

            await curr.execute(
                query,
            )

            todos = await curr.fetchall()

            return todos


async def select_todo(db: AsyncConnectionPool, id: int, member_id: int) -> Todo | None:
    """
    Find exactly one todo task that matches id and member_id
    in todos list in PostgreSQL database

    Return a Todo object or None
    """
    async with db.connection() as conn:
        async with conn.cursor(row_factory=class_row(Todo)) as curr:
            query = sql.SQL(
                """
                SELECT id, complete, due, task
                    FROM todos
                    WHERE id = (%s) AND member_id = (%s)
                """
            )

            await curr.execute(query, (id, member_id))

            member = await curr.fetchone()

            return None if member is None else member


async def insert_todo(
    db: AsyncConnectionPool,
    member_id: int,
    task: str,
    complete: bool,
    due: datetime | None,
) -> Todo | None:
    """
    Insert a todo into the database

    Return the inserted todo object or None
    """
    async with db.connection() as conn:
        async with conn.cursor(row_factory=class_row(Todo)) as curr:
            query = sql.SQL(
                """
                INSERT INTO todos (complete, due, member_id, task)
                    VALUES ((%s), (%s), (%s), (%s))
                    RETURNING id, complete, due, task
                """
            )
            await curr.execute(
                query,
                (
                    complete,
                    due,
                    member_id,
                    task,
                ),
            )

            todo = await curr.fetchone()

            return None if todo is None else todo


async def update_todo(
    db: AsyncConnectionPool,
    id: int,
    member_id: int,
    task: str,
    complete: bool,
    due: datetime | None,
) -> Todo | None:
    """
    Update a todo in the database

    Return the updated todo object or None
    """
    async with db.connection() as conn:
        async with conn.cursor(row_factory=class_row(Todo)) as curr:
            query = sql.SQL(
                """
                UPDATE todos
                    SET complete = (%s), due = (%s), task = (%s)
                    WHERE id = (%s) AND member_id = (%s)
                    RETURNING id, complete, due, task
                """
            )
            await curr.execute(
                query,
                (
                    complete,
                    due,
                    task,
                    id,
                    member_id,
                ),
            )

            todo = await curr.fetchone()

            return None if todo is None else todo


async def delete_todo(db: AsyncConnectionPool, id: int, member_id: int) -> bool:
    """
    Delete a todo with an id and its member_id. Note that you can not delete other users' todos

    Return True if deletion is successful and False otherwise
    """
    async with db.connection() as conn:
        async with conn.cursor(row_factory=class_row(Todo)) as curr:
            query = sql.SQL(
                """
                DELETE FROM todos 
                    WHERE id = (%s) AND member_id = (%s)
                """
            )
            await curr.execute(
                query,
                (
                    id,
                    member_id,
                ),
            )

            return curr.rowcount > 0
